@@ -4,7 +4,19 @@
 #include <lib/stringinfo.h>
 #include <access/htup_details.h>
 #include <access/tupmacs.h>
+#if PG_VERSION_NUM >= 140000
+#ifdef USE_LZ4
+#include <lz4.h>
+#endif
+#include <access/toast_compression.h>
+#include <access/toast_internals.h>
+#endif
+#if PG_VERSION_NUM >= 130000
+#include <access/detoast.h>
+#include <access/heaptoast.h>
+#else
 #include <access/tuptoaster.h>
+#endif
 #include <datatype/timestamp.h>
 #include <common/pg_lzcompress.h>
 #include <string.h>
@@ -23,7 +35,24 @@ ReadStringFromToast(const char *buffer,
  * Utilities for manipulation of header information for compressed
  * toast entries.
  */
-#define TOAST_COMPRESS_RAWSIZE(ptr) (*(uint32 *) ptr)
+#if PG_VERSION_NUM < 140000
+/*
+ * These macros define the "saved size" portion of va_extinfo.  Its remaining
+ * two high-order bits identify the compression method.
+ * Before std14 only pglz compression method existed (with 00 bits).
+ */
+#define VARLENA_EXTSIZE_BITS   30
+#define VARLENA_EXTSIZE_MASK   ((1U << VARLENA_EXTSIZE_BITS) - 1)
+#define VARDATA_COMPRESSED_GET_COMPRESS_METHOD(ptr) ((*((uint32 *)ptr + 1)) >> VARLENA_EXTSIZE_BITS)
+typedef enum ToastCompressionId
+{
+	TOAST_PGLZ_COMPRESSION_ID = 0,
+	TOAST_LZ4_COMPRESSION_ID = 1,
+	TOAST_INVALID_COMPRESSION_ID = 2
+} ToastCompressionId;
+#endif
+#define TOAST_COMPRESS_RAWSIZE(ptr) ((*(uint32 *) ptr) & VARLENA_EXTSIZE_MASK)
+#define TOAST_COMPRESS_RAWMETHOD(ptr) ((*(uint32 *) ptr) >> VARLENA_EXTSIZE_BITS)
 #define TOAST_COMPRESS_RAWDATA(ptr) (ptr + sizeof(uint32))
 #define TOAST_COMPRESS_HEADER_SIZE (sizeof(uint32))
 
@@ -517,7 +546,7 @@ decode_bigint(const char *buffer, unsigned int buff_size, unsigned int *out_size
 	if (buff_size < sizeof(int64))
 		return -2;
 
-	CopyAppendFmt("%ld", *(int64 *) buffer);
+	CopyAppendFmt(INT64_FORMAT, *(int64 *) buffer);
 	*out_size = sizeof(int64) + delta;
 	return 0;
 }
@@ -544,7 +573,7 @@ decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 	timestamp_sec = timestamp / 1000000;
 	*out_size = sizeof(int64) + delta;
 
-	CopyAppendFmt("%02ld:%02ld:%02ld.%06ld",
+	CopyAppendFmt("%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d",
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000);
 
@@ -577,7 +606,7 @@ decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size
 	tz_min = -(tz_sec / 60);
 	*out_size = sizeof(int64) + sizeof(int32) + delta;
 
-	CopyAppendFmt("%02ld:%02ld:%02ld.%06ld%c%02d:%02d",
+	CopyAppendFmt("%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%c%02d:%02d",
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000, (tz_min > 0 ? '+' : '-'), abs(tz_min / 60), abs(tz_min % 60));
 
@@ -655,7 +684,7 @@ decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_s
 	j2date(jd, &year, &month, &day);
 	timestamp_sec = timestamp / 1000000;
 
-	CopyAppendFmt("%04d-%02d-%02d %02ld:%02ld:%02ld.%06ld%s",
+	CopyAppendFmt("%04d-%02d-%02d %02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%s",
 				  (year <= 0) ? -year + 1 : year, month, day,
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000,
@@ -879,9 +908,17 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 		/*
 		 * xxxxxx10 4-byte length word, aligned, *compressed* data (up to 1G)
 		 */
-		int			decompress_ret;
-		uint32		len = VARSIZE_4B(buffer);
-		uint32		decompressed_len = VARRAWSIZE_4B_C(buffer);
+		int						decompress_ret;
+		uint32					len = VARSIZE_4B(buffer);
+		uint32					decompressed_len = 0;
+		ToastCompressionId		cmid;
+
+#if PG_VERSION_NUM >= 140000
+		decompressed_len = VARDATA_COMPRESSED_GET_EXTSIZE(buffer);
+#else
+		decompressed_len = VARRAWSIZE_4B_C(buffer);
+#endif
+
 
 		if (len > buff_size)
 			return -1;
@@ -897,8 +934,32 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 			return 0;
 		}
 
-		decompress_ret = pglz_decompress(VARDATA_4B_C(buffer), len - 2 * sizeof(uint32),
-										 decompress_tmp_buff, decompressed_len);
+		cmid = VARDATA_COMPRESSED_GET_COMPRESS_METHOD(buffer);
+		switch(cmid)
+		{
+			case TOAST_PGLZ_COMPRESSION_ID:
+				decompress_ret = pglz_decompress(VARDATA_4B_C(buffer), len - 2 * sizeof(uint32),
+												 decompress_tmp_buff, decompressed_len
+#if PG_VERSION_NUM >= 120000
+												 , true
+#endif
+												 );
+				break;
+			case TOAST_LZ4_COMPRESSION_ID:
+#ifdef USE_LZ4
+				decompress_ret = LZ4_decompress_safe(VARDATA_4B_C(buffer), decompress_tmp_buff,
+													 len - 2 * sizeof(uint32), decompressed_len);
+				break;
+#else
+				printf("Error: compression method lz4 not supported.\n");
+				printf("Try to rebuild pg_filedump for PostgreSQL server of version 14+ with --with-lz4 option.\n");
+				return -2;
+#endif
+			default:
+				decompress_ret = -1;
+				break;
+		}
+
 		if ((decompress_ret != decompressed_len) || (decompress_ret < 0))
 		{
 			printf("WARNING: Unable to decompress a string. Data is corrupted.\n");
@@ -972,24 +1033,50 @@ FormatDecode(const char *tupleData, unsigned int tupleSize)
 	CopyFlush();
 }
 
-static int DumpCompressedString(const char *data, int32 decompressed_size)
+static int DumpCompressedString(const char *data, int32 compressed_size)
 {
-	int		decompress_ret;
-	char   *decompress_tmp_buff = malloc(TOAST_COMPRESS_RAWSIZE(data));
+	int						decompress_ret;
+	char				   *decompress_tmp_buff = malloc(TOAST_COMPRESS_RAWSIZE(data));
+	ToastCompressionId		cmid;
+
+	cmid = TOAST_COMPRESS_RAWMETHOD(data);
+	switch(cmid)
+	{
+		case TOAST_PGLZ_COMPRESSION_ID:
+			decompress_ret = pglz_decompress(TOAST_COMPRESS_RAWDATA(data),
+											 compressed_size - TOAST_COMPRESS_HEADER_SIZE,
+											 decompress_tmp_buff, TOAST_COMPRESS_RAWSIZE(data)
+#if PG_VERSION_NUM >= 120000
+											 , true
+#endif
+											 );
+			break;
+		case TOAST_LZ4_COMPRESSION_ID:
+#ifdef USE_LZ4
+			decompress_ret = LZ4_decompress_safe(TOAST_COMPRESS_RAWDATA(data),
+												 decompress_tmp_buff, compressed_size - TOAST_COMPRESS_HEADER_SIZE,
+												 TOAST_COMPRESS_RAWSIZE(data));
+			break;
+#else
+			printf("Error: compression method lz4 not supported.\n");
+			printf("Try to rebuild pg_filedump for PostgreSQL server of version 14+ with --with-lz4 option.\n");
+			return -2;
+#endif
+		default:
+			decompress_ret = -1;
+			break;
+	}
 
-	decompress_ret = pglz_decompress(TOAST_COMPRESS_RAWDATA(data),
-			decompressed_size - TOAST_COMPRESS_HEADER_SIZE,
-			decompress_tmp_buff, TOAST_COMPRESS_RAWSIZE(data));
 	if ((decompress_ret != TOAST_COMPRESS_RAWSIZE(data)) ||
 			(decompress_ret < 0))
 	{
 		printf("WARNING: Unable to decompress a string. Data is corrupted.\n");
 		printf("Returned %d while expected %d.\n", decompress_ret,
-				decompressed_size);
+				TOAST_COMPRESS_RAWSIZE(data));
 	}
 	else
 	{
-		CopyAppendEncode(decompress_tmp_buff, *((uint32 *)data));
+		CopyAppendEncode(decompress_tmp_buff, decompress_ret);
 	}
 
 	free(decompress_tmp_buff);
@@ -1022,15 +1109,21 @@ ReadStringFromToast(const char *buffer,
 		unsigned int control_options = 0;
 
 		VARATT_EXTERNAL_GET_POINTER(toast_ptr, buffer);
+
+		/* Extract TOASTed value */
+#if PG_VERSION_NUM >= 140000
+		toast_ext_size = VARATT_EXTERNAL_GET_EXTSIZE(toast_ptr);
+#else
+		toast_ext_size = toast_ptr.va_extsize;
+#endif
+
 		printf("  TOAST value. Raw size: %8d, external size: %8d, "
 				"value id: %6d, toast relation id: %6d\n",
 				toast_ptr.va_rawsize,
-				toast_ptr.va_extsize,
+				toast_ext_size,
 				toast_ptr.va_valueid,
 				toast_ptr.va_toastrelid);
 
-		/* Extract TOASTed value */
-		toast_ext_size = toast_ptr.va_extsize;
 		num_chunks = (toast_ext_size - 1) / TOAST_MAX_CHUNK_SIZE + 1;
 		printf("  Number of chunks: %d\n", num_chunks);
 
@@ -1061,7 +1154,7 @@ ReadStringFromToast(const char *buffer,
 					-1, /* no end block */
 					true, /* is toast relation */
 					toast_ptr.va_valueid,
-					toast_ptr.va_extsize,
+					toast_ext_size,
 					toast_data);
 
 			if (result == 0)
