@@ -3,7 +3,7 @@
  *				   formatting heap (data), index and control files.
  *
  * Copyright (c) 2002-2010 Red Hat, Inc.
- * Copyright (c) 2011-2018, PostgreSQL Global Development Group
+ * Copyright (c) 2011-2021, PostgreSQL Global Development Group
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -83,9 +83,30 @@ static unsigned int bytesToFormat = 0;
 /* Block version number */
 static unsigned int blockVersion = 0;
 
+/* Flag to indicate pg_filenode.map file */
+static bool isRelMapFile = false;
+
 /* Program exit code */
 static int	exitCode = 0;
 
+/* Relmapper structs */
+typedef struct RelMapping
+{
+  Oid     mapoid;     /* OID of a catalog */
+  Oid     mapfilenode;  /* its filenode number */
+} RelMapping;
+
+/* crc and pad are ignored here, even though they are
+ * present in the backend code.  We assume that anyone
+ * seeking to inspect the contents of pg_filenode.map
+ * probably have a corrupted or non-functional cluster */
+typedef struct RelMapFile
+{
+  int32   magic;      /* always RELMAPPER_FILEMAGIC */
+  int32   num_mappings; /* number of valid RelMapping entries */
+  RelMapping  mappings[FLEXIBLE_ARRAY_MEMBER];
+} RelMapFile;
+
 /*
  * Function Prototypes
  */
@@ -127,6 +148,7 @@ static void FormatControl(char *buffer);
 static void FormatBinary(char *buffer,
 		unsigned int numBytes, unsigned int startIndex);
 static void DumpBinaryBlock(char *buffer);
+static int PrintRelMappings(void);
 
 
 /* Send properly formed usage information to the user. */
@@ -137,7 +159,7 @@ DisplayOptions(unsigned int validOptions)
 		printf
 			("\nVersion %s (for %s)"
 			 "\nCopyright (c) 2002-2010 Red Hat, Inc."
-		  "\nCopyright (c) 2011-2018, PostgreSQL Global Development Group\n",
+		  "\nCopyright (c) 2011-2021, PostgreSQL Global Development Group\n",
 			 FD_VERSION, FD_PG_VERSION);
 
 	printf
@@ -179,6 +201,9 @@ DisplayOptions(unsigned int validOptions)
 		 "  -c  Interpret the file listed as a control file\n"
 		 "  -f  Display formatted content dump along with interpretation\n"
 		 "  -S  Force block size to [blocksize]\n"
+		 "Additional functions:\n"
+		 "  -m  Interpret file as pg_filenode.map file and print contents (all\n"
+		 "      other options will be ignored)\n" 
 		 "\nReport bugs to <pgsql-bugs@postgresql.org>\n");
 }
 
@@ -515,6 +540,11 @@ ConsumeOptions(int numOptions, char **options)
 						SET_OPTION(blockOptions, BLOCK_CHECKSUMS, 'k');
 						break;
 
+						/* Treat file as pg_filenode.map file */
+					case 'm':
+						isRelMapFile = true;
+						break;
+
 						/* Display old values. Ignore Xmax */
 					case 'o':
 						SET_OPTION(blockOptions, BLOCK_IGNORE_OLD, 'o');
@@ -659,6 +689,13 @@ GetBlockSize(FILE *fp)
 		exitCode = 1;
 	}
 
+	if (localSize == 0)
+	{
+		printf("Notice: Block size determined from reading block 0 is zero, using default %d instead.\n", BLCKSZ);
+		printf("Hint: Use -S <size> to specify the size manually.\n");
+		localSize = BLCKSZ;
+	}
+
 	return (localSize);
 }
 
@@ -779,7 +816,6 @@ CreateDumpFileHeader(int numOptions, char **options)
 {
 	unsigned int x;
 	char		optionBuffer[52] = "\0";
-	time_t		rightNow = time(NULL);
 
 	/* Iterate through the options and cache them.
 	 * The maximum we can display is 50 option characters + spaces. */
@@ -788,19 +824,18 @@ CreateDumpFileHeader(int numOptions, char **options)
 		if ((strlen(optionBuffer) + strlen(options[x])) > 50)
 			break;
 		strcat(optionBuffer, options[x]);
-		strcat(optionBuffer, " ");
+		if (x < numOptions - 2)
+			strcat(optionBuffer, " ");
 	}
 
 	printf
 		("\n*******************************************************************\n"
-		 "* PostgreSQL File/Block Formatted Dump Utility - Version %s\n"
+		 "* PostgreSQL File/Block Formatted Dump Utility\n"
 		 "*\n"
 		 "* File: %s\n"
-		 "* Options used: %s\n*\n"
-		 "* Dump created on: %s"
+		 "* Options used: %s\n"
 		 "*******************************************************************\n",
-		 FD_VERSION, fileName, (strlen(optionBuffer)) ? optionBuffer : "None",
-		 ctime(&rightNow));
+		 fileName, (strlen(optionBuffer)) ? optionBuffer : "None");
 }
 
 /*	Dump out a formatted block header for the requested block */
@@ -969,7 +1004,7 @@ FormatItemBlock(char *buffer,
 		return;
 
 	if (!isToast || verbose)
-		printf("%s<Data> ------ \n", indent);
+		printf("%s<Data> -----\n", indent);
 
 	/* Loop through the items on the block.  Check if the block is
 	 * empty and has a sensible item array listed before running
@@ -1239,7 +1274,11 @@ FormatItem(char *buffer, unsigned int numBytes, unsigned int startIndex,
 	else if (formatAs == ITEM_SPG_LEAF)
 	{
 		/* It is an SpGistLeafTuple item, so dump the index header */
+#if PG_VERSION_NUM >= 140000
+		if (numBytes < SGLTHDRSZ(SGLT_GET_HASNULLMASK((SpGistLeafTuple) &(buffer[startIndex]))))
+#else
 		if (numBytes < SGLTHDRSZ)
+#endif
 		{
 			if (numBytes)
 			{
@@ -1253,7 +1292,11 @@ FormatItem(char *buffer, unsigned int numBytes, unsigned int startIndex,
 
 			printf("  State: %s  nextOffset: %u  Block Id: %u  linp Index: %u\n\n",
 				   spgist_tupstates[itup->tupstate],
+#if PG_VERSION_NUM >= 140000
+				   SGLT_GET_NEXTOFFSET(itup),
+#else
 				   itup->nextOffset,
+#endif
 				   ((uint32) ((itup->heapPtr.ip_blkid.bi_hi << 16) |
 							  (uint16) itup->heapPtr.ip_blkid.bi_lo)),
 				   itup->heapPtr.ip_posid);
@@ -1307,9 +1350,11 @@ FormatItem(char *buffer, unsigned int numBytes, unsigned int startIndex,
 				   HeapTupleHeaderGetRawXmax(htup),
 				   HeapTupleHeaderGetRawCommandId(htup));
 
+#if PG_VERSION_NUM < 120000
 			if (infoMask & HEAP_HASOID)
 				printf("  OID: %u",
 					   HeapTupleHeaderGetOid(htup));
+#endif
 
 			printf("\n"
 				   "  Block Id: %u  linp Index: %u   Attributes: %d   Size: %d\n",
@@ -1327,8 +1372,10 @@ FormatItem(char *buffer, unsigned int numBytes, unsigned int startIndex,
 				strcat(flagString, "HASVARWIDTH|");
 			if (infoMask & HEAP_HASEXTERNAL)
 				strcat(flagString, "HASEXTERNAL|");
+#if PG_VERSION_NUM < 120000
 			if (infoMask & HEAP_HASOID)
 				strcat(flagString, "HASOID|");
+#endif
 			if (infoMask & HEAP_XMAX_KEYSHR_LOCK)
 				strcat(flagString, "XMAX_KEYSHR_LOCK|");
 			if (infoMask & HEAP_COMBOCID)
@@ -1373,8 +1420,10 @@ FormatItem(char *buffer, unsigned int numBytes, unsigned int startIndex,
 			else
 				bitmapLength = 0;
 
+#if PG_VERSION_NUM < 120000
 			if (infoMask & HEAP_HASOID)
 				oidLength += sizeof(Oid);
+#endif
 
 			computedLength =
 				MAXALIGN(localBitOffset + bitmapLength + oidLength);
@@ -1463,7 +1512,11 @@ FormatSpecial(char *buffer)
 					   btreeSection->btpo_prev, btreeSection->btpo_next,
 					   (btreeSection->
 						btpo_flags & BTP_DELETED) ? "Next XID" : "Level",
+#if PG_VERSION_NUM >= 140000
+					   btreeSection->btpo_level,
+#else
 					   btreeSection->btpo.level,
+#endif
 					   btreeSection->btpo_cycleid);
 			}
 			break;
@@ -1767,7 +1820,15 @@ FormatControl(char *buffer)
 #endif
 			   (uint32) (checkPoint->redo >> 32), (uint32) checkPoint->redo,
 			   checkPoint->ThisTimeLineID,
+#if PG_VERSION_NUM < 120000
 			   checkPoint->nextXidEpoch, checkPoint->nextXid,
+#elif PG_VERSION_NUM < 140000
+			   EpochFromFullTransactionId(checkPoint->nextFullXid),
+			   XidFromFullTransactionId(checkPoint->nextFullXid),
+#else
+			   EpochFromFullTransactionId(checkPoint->nextXid),
+			   XidFromFullTransactionId(checkPoint->nextXid),
+#endif
 			   checkPoint->nextOid,
 			   checkPoint->nextMulti, checkPoint->nextMultiOffset,
 			   ctime(&cp_time),
@@ -1981,6 +2042,69 @@ DumpFileContents(unsigned int blockOptions,
 	return result;
 }
 
+int
+PrintRelMappings(void)
+{
+	// For storing ingested data
+	char charbuf[RELMAPPER_FILESIZE];
+	RelMapFile *map;
+	RelMapping *mappings;
+	RelMapping m;
+	int bytesRead;
+
+	// For confirming Magic Number correctness
+	char m1[RELMAPPER_MAGICSIZE];
+	char m2[RELMAPPER_MAGICSIZE];
+	int magic_ref = RELMAPPER_FILEMAGIC;
+	int magic_val;
+	int num_loops;
+
+	// Read in the file
+	rewind(fp); // Make sure to start from the beginning
+	bytesRead = fread(charbuf,1,RELMAPPER_FILESIZE,fp);
+	if ( bytesRead != RELMAPPER_FILESIZE ) {
+		printf("Read %d bytes, expected %d\n", bytesRead, RELMAPPER_FILESIZE);
+		return 0;
+	}
+
+	// Convert to RelMapFile type for usability
+	map = (RelMapFile *) charbuf;
+
+
+	// Check and print Magic Number correctness
+	printf("Magic Number: 0x%x",map->magic);
+	magic_val = map->magic;
+
+	memcpy(m1,&magic_ref,RELMAPPER_MAGICSIZE);
+	memcpy(m2,&magic_val,RELMAPPER_MAGICSIZE);
+	if ( memcmp(m1,m2,RELMAPPER_MAGICSIZE) == 0 ) {
+		printf(" (CORRECT)\n");
+	} else {
+		printf(" (INCORRECT)\n");
+	}
+
+	// Print Mappings
+	printf("Num Mappings: %d\n",map->num_mappings);
+	printf("Detailed Mappings list:\n");
+	mappings = map->mappings;
+
+	// Limit number of mappings as per MAX_MAPPINGS
+	num_loops = map->num_mappings;
+	if ( map->num_mappings > MAX_MAPPINGS ) {
+		num_loops = MAX_MAPPINGS;
+		printf("  NOTE: listing has been limited to the first %d mappings\n", MAX_MAPPINGS);
+		printf("        (perhaps your file is not a valid pg_filenode.map file?)\n");
+	}
+
+	for (int i=0; i < num_loops; i++) {
+		m = mappings[i];
+		printf("OID: %u\tFilenode: %u\n",
+			m.mapoid,
+			m.mapfilenode);
+	}
+	return 1;
+}
+
 /* Consume the options and iterate through the given file, formatting as
  * requested. */
 int
@@ -1995,6 +2119,11 @@ main(int argv, char **argc)
 	 * where encountered */
 	if (validOptions != OPT_RC_VALID)
 		DisplayOptions(validOptions);
+	else if (isRelMapFile)
+	{
+		CreateDumpFileHeader(argv, argc);
+		exitCode = PrintRelMappings();
+	}
 	else
 	{
 		/* Don't dump the header if we're dumping binary pages */
